from datetime import datetime, timedelta
from json import dumps
from Microsoft.Finance.Common.Spark import Spark
from pyspark.sql.session import SparkSession
from pyspark.sql.dataframe import DataFrame
from unittest import TestSuite, TextTestRunner, TestLoader

class BaseTestSuite(object):

    #region Constructor

    def __init__(self, sparkSession: SparkSession, databaseName: str, tableName: str) -> None:
        self.__spark = Spark(sparkSession)
        self.__databaseName = databaseName
        self.__tableName = tableName
        self.__customEvents = []
        self.__tableCount = None
        self.__tableDataFrame = None
        self.__fieldInfo = None
        self.__duration = None
        self.__testSuite = TestSuite()
        self.__testRunner = TextTestRunner()

    #endregion

    #region Properties

    @property
    def TestRunner(self) -> TextTestRunner:
        return self.__testRunner
    
    @property
    def SparkUtility(self) -> Spark:
        return self.__spark

    @property
    def DatabaseName(self) -> str:
        return self.__databaseName

    @property
    def TableName(self) -> str:
        return self.__tableName

    @property
    def CustomEvents(self) -> {}:
        return self.__customEvents

    @property
    def TestSuite(self) -> TestSuite:
        return self.__testSuite

    @property
    def TableCount(self) -> int:
        if not self.__tableCount:
            self.__tableCount = self.__spark.count_table_or_view(self.__databaseName, self.__tableName)
        return self.__tableCount

    @property
    def TableDataFrame(self) -> DataFrame:
        if not self.__tableDataFrame:
            self.__tableDataFrame = self.__spark.get_table_or_view(self.__databaseName, self.__tableName)
        return self.__tableDataFrame

    @property
    def FieldInfo(self) -> {(str, type)}:
        if not self.__fieldInfo:
            self.__fieldInfo = self.__spark.get_field_info(self.TableDataFrame)
        return self.__fieldInfo

    @property 
    def SuiteDurationSeconds(self) -> float:
        if not __customEvents:
            return 0.0
        else:
            if not self.__duration:
                self.__duration = sum([event.EventDurationSeconds for event in self.__customEvents])
            return self.__duration.total_seconds()

    @property
    def SuiteDurationMilliseconds(self) -> float:
        return self.SuiteDurationSeconds * 1000

    #endregion

    #region Class Methods

    def __str__(self) -> str:
        s = dumps(self.__dict__)
        return s

    def dump_events(self) -> None:

        for event in self.CustomEvents:
            print("Event Name: {}\nEvent Details: {}".format(event.TestName, str(event)))
    
    def run_tests(self) -> None:

        print("TEST RESULTS")

        self.TestRunner.run(self.TestSuite)

        print("\nEVENT DETAILS")

        self.dump_events()

    #endregion



from Microsoft.Finance.Common.Spark import Spark
from Microsoft.Finance.Common.Utilities import Utilities
from Microsoft.Finance.DataStandards.Utilities.TestEvent import TestEvent
from Microsoft.Finance.DataStandards.Utilities.TestUtilities import TestUtilities
from pyspark.sql.dataframe import DataFrame

import unittest

class FactTestCase(unittest.TestCase):

    #region Constructor

    def __init__(self, sparkUtility: Spark, tableDataFrame: DataFrame, fieldInfo: {(str, type)}, linkedDimensionTables: [(str, str)], databaseName: str, tableName: str, tableCount: int, customEvents: [TestEvent], excludeFromAggregation: [str] = [], methodName: str = "runTest"):
        super(FactTestCase, self).__init__(methodName)
        self.__spark = sparkUtility
        self.__tableDataFrame = tableDataFrame
        self.__fieldInfo = fieldInfo
        self.__linkedDimensionTables = linkedDimensionTables
        self.__databaseName = databaseName
        self.__tableName = tableName
        self.__tableCount = tableCount
        self.__customEvents = customEvents
        self.__excludeFromAggregation = excludeFromAggregation

    #endregion

    #region Class Methods

    def test_aggregation_efficiency(self) -> None:
        
        '''
        desc
            Conventions for fact tables
                e. Fact columns should be aggregated wherever possible
        '''

        event = TestEvent(Utilities.get_function_caller_name())

        eventCode1 = "fact-columns"
        eventCode2 = "group-by-columns"
        eventCode3 = "custom-query"
        eventCode4 = "custom-count"

        factColumns = []
        groupByColumns = []

        event.start()

        for columnName, columnType in self.__fieldInfo:
            if columnName in self.__excludeFromAggregation or not TestUtilities.check_fact_type(columnType):
                groupByColumns.append(columnName)
            else:
                factColumns.append(columnName)
        
        factColumnAggregations = ["sum({})".format(columnName) for columnName in factColumns]
        
        factColumnString = ", ".join(factColumnAggregations)
        groupByColumnString = ", ".join(groupByColumns)

        customQuery = "select {}, {} from {}.{} group by {}".format(groupByColumnString, factColumnString, self.__databaseName, self.__tableName, groupByColumnString)
        customCount = self.__spark.count_query(customQuery)

        event.end()

        event.add_note(eventCode1, factColumns)
        event.add_note(eventCode2, groupByColumns)
        event.add_note(eventCode3, customQuery)
        event.add_note(eventCode4, customCount)

        self.__customEvents.append(event)

        with self.subTest():
            self.assertLessEqual(self.__tableCount, customCount)

    def test_column_categories(self) -> None:
        
        '''
        desc
            Conventions for fact tables
                c. Output table should only contain dimension Ids, fact columns and in some cases, degenerate attributes. Degenerate attributes should be added only where it is needed
        '''

        event = TestEvent(Utilities.get_function_caller_name())

        eventCode1 = "dimension-keys"
        eventcode2 = "degenerate-attributes"
        eventCode3 = "fact-columns"
        eventCode4 = "bad-columns"

        dimensionKeys = []
        degenerateAttributes = []
        facts = []
        badColumns = []

        event.start()

        for columnName, columnType in self.__fieldInfo:
            if TestUtilities.check_dimension_key_name(columnName):
                if TestUtilities.check_dimension_key_type(columnType):
                    dimensionKeys.append((columnName, columnType))
                else:
                    badColumns.append((columnName, columnType))
            elif TestUtilities.check_fact_type(columnType):
                facts.append((columnName, columnType))
            else:
                degenerateAttributes.append((columnName, columnType))

        event.end()

        event.add_note(eventCode1, dimensionKeys)
        event.add_note(eventCode2, degenerateAttributes)
        event.add_note(eventCode3, facts)
        event.add_note(eventCode4, badColumns)

        self.__customEvents.append(event)
        
        with self.subTest():
            self.assertEqual(len(badColumns), 0)
        
        with self.subTest():
            self.assertGreater(len(facts), 0)
        
        with self.subTest():
            self.assertGreater(len(dimensionKeys), 0)

    def test_column_order(self) -> None:
        
        '''
        desc
            Conventions for fact tables
                i. Column sequencing should have all dimension Id columns first, followed by degenerate attributes and facts at the end
        '''

        event = TestEvent(Utilities.get_function_caller_name())

        eventCode1 = "out-of-order-columns"
        eventCode2 = "index-of-last-dimension-key"
        eventCode3 = "index-of-last-degerate-attribute"

        outOfOrderColumns = []
        indexOfLastDimension = -float("inf")
        indexOfLastDegenerateAttribute = -float("inf")

        event.start()

        N = len(self.__fieldInfo)

        for i in range(N):
            if i < N - 1:
                if TestUtilities.check_dimension_key_name_and_type(self.__fieldInfo[i][0], self.__fieldInfo[i][1]):
                    if not TestUtilities.check_dimension_key_name_and_type(self.__fieldInfo[i + 1][0], self.__fieldInfo[i + 1][1]):
                        indexOfLastDimension = i
                        break
        
        degenerateAttributeStart = indexOfLastDimension + 1 if indexOfLastDimension > -float("inf") else 0

        for i in range(degenerateAttributeStart, N):
            if i < N - 1:
                if not TestUtilities.check_fact_type(self.__fieldInfo[i][1]) and TestUtilities.check_fact_type(self.__fieldInfo[i + 1][1]):
                    indexOfLastDegenerateAttribute = i
                    break
        
        factStart = indexOfLastDegenerateAttribute + 1 if indexOfLastDegenerateAttribute > -float("inf") else degenerateAttributeStart

        if indexOfLastDimension > -float("inf"):
            for i in range(indexOfLastDimension + 1):
                columnName = self.__fieldInfo[i][0]
                columnType = self.__fieldInfo[i][1]
                if not TestUtilities.check_dimension_key_name_and_type(columnName, columnType):
                    outOfOrderColumns.append((columnName, columnType))
        
        if indexOfLastDegenerateAttribute > float("inf"):
            for i in range(degenerateAttributeStart, indexOfLastDegenerateAttribute + 1):
                columnName = self.__fieldInfo[i][0]
                columnType = self.__fieldInfo[i][0]
                if TestUtilities.check_dimension_key_name(columnName) or TestUtilities.check_dimension_key_type(columnName):
                    outOfOrderColumns.append((columnName, columnType))
        
        for i in range(factStart, N):
            columnName = self.__fieldInfo[i][0]
            columnType = self.__fieldInfo[i][1]
            if not TestUtilities.check_fact_type(columnType):
                outOfOrderColumns.append((columnName, columnType))

        event.end()

        event.add_note(eventCode1, outOfOrderColumns)
        event.add_note(eventCode2, indexOfLastDimension)
        event.add_note(eventCode3, indexOfLastDegenerateAttribute)

        self.__customEvents.append(event)
        
        self.assertEqual(len(outOfOrderColumns), 0)

    def test_dimension_id_value_not_null(self) -> None:
        
        '''
        desc
            Conventions for fact tables
                d. Dimension Id columns included in fact tables should adhere to following:
                    d.1) Dimension Id column in fact should not be NULL
        '''

        event = TestEvent(Utilities.get_function_caller_name())

        eventCode1 = "dimension-keys-with-nulls"

        dimensionsWithNulls = []

        event.start()

        for columnName, columnType in self.__fieldInfo:
            if TestUtilities.check_dimension_key_name_and_type(columnName, columnType):

                if not tableDf:
                    tableDf = self.__spark.get_table_or_view(self.__databaseName, self.__tableName)

                uniqueValues = self.__spark.get_distinct_column_values_from_df(self.__tableDataFrame, columnName)

                for val in uniqueValues:
                    if val == None:
                        dimensionsWithNulls.append((columnName, columnType))
                        break
        event.end()

        event.add_note(eventCode1, dimensionsWithNulls)
        
        self.assertEqual(len(dimensionsWithNulls), 0)

    def test_dimension_id_values_in_dimension(self) -> None:
        
        '''
        desc
            Conventions for fact tables
                d. Dimension Id columns included in fact tables should adhere to following:
                    d.2) Values in dimension Id column of fact should be present in actual dimension
        '''

        event = TestEvent(Utilities.get_function_caller_name())

        eventCode1 = "keys-with-bad-values"

        keysWithBadValues = []
        dimensionFieldInfo = {}

        event.start()

        dimensionKeys = [column[0] for column in self.__fieldInfo if TestUtilities.check_dimension_key_name_and_type(column[0], column[1])]

        for dimensionInfo in self.__linkedDimensionTables:
            dimensionDatabaseName = dimensionInfo[0]
            dimensionTableName = dimensionInfo[1]
            dimensionFieldInfo[(dimensionDatabaseName, dimensionTableName)] = self.__spark.get_field_info_table_or_view(dimensionDatabaseName, dimensionTableName)

        for key in dimensionKeys:
            for dimInfo, schema in dimensionInfo.items():
                
                dimensionColumns = {column[0] for column in schema}
                
                if key in dimensionColumns:
                    
                    dimensionDatabaseName = dimInfo[0]
                    dimensionTableName = dimInfo[1]
                    
                    keyValues = self.__spark.get_distinct_column_values_from_df(self.__tableDataFrame, key)
                    dimensionKeyValues = self.__spark.get_distinct_column_values(self.__sparkSession, dimensionDatabaseName, dimensionTableName, key)
                    
                    diff = keyValues - dimensionKeyValues
                    
                    if diff:
                        keysWithBadValues.append((dimensionTableName, key, diff))
        
        event.end()

        event.add_note(eventCode1, keysWithBadValues)

        self.assertEqual(len(keysWithBadValues), 0)

    def test_key_column_name_in_dimension(self) -> None:
        
        '''
        desc
            Conventions for fact tables
                d. Dimension Id columns included in fact tables should adhere to following:
                    d.4) Name of Dimension Id column should match to the actual column name in dimension table except in cases where the fact table relates to same dimension using multiple columns. In such cases, the relevant name should be added before DimensionId. E.g. DIM_StartDateId, DIM_EndDateId
        '''

        event = TestEvent(Utilities.get_function_caller_name())

        eventCode1 = "keys-not-in-dimension"

        keysNotInDimension = []
        keysInDimensions = []
        dimensionFieldInfo = {}

        event.start()

        for dimensionInfo in self.__linkedDimensionTables:
            dimensionDatabaseName = dimensionInfo[0]
            dimensionTableName = dimensionInfo[1]
            dimensionFieldInfo[dimensionTableName] = self.__spark.get_field_info(dimensionDatabaseName, dimensionTableName)

        keyColumns = [column[0] for column in self.__fieldInfo if TestUtilities.check_dimension_key_name_and_type(column[0], column[1])]
        
        for columnName in keyColumns:
            inDimension = False
            for dimensionName, dimensionSchema in dimensionFieldInfo.items():
                dimensionColumns = {column[0] for column in dimensionSchema}
                if columnName in dimensionColumns:
                    inDimension = True
                    keysInDimensions.append((columnName, dimensionName))
                    break
            
            if not inDimension:
                keysNotInDimension.append((columnName))

        event.end()

        event.add_note(eventCode1, keysNotInDimension)
        
        with self.subTest():
            self.assertEquals(len(keysNotInDimension), 0)
        
        with self.subTest():
            self.assertGreater(len(keysInDimensions), 0)

    #endregion


from Microsoft.Finance.DataStandards.UnitTest.TestCase.BridgeTestCase import BridgeTestCase
from Microsoft.Finance.DataStandards.UnitTest.TestCase.CommonTestCase import CommonTestCase
from Microsoft.Finance.DataStandards.UnitTest.TestSuite.BaseTestSuite import BaseTestSuite
from pyspark.sql.session import SparkSession

class BridgeTestSuite(BaseTestSuite):

    #region Constructor

    def __init__(self, sparkSession: SparkSession, databaseName: str, tableName: str):
        super(BridgeTestSuite, self).__init__(sparkSession, databaseName, tableName)

    #endregion

    #region Class Methods

    def queue_tests(self) -> None:
        bridgeTestCase = BridgeTestCase(self.FieldInfo, self.CustomEvents)
        commonTestCase = CommonTestCase(self.TableDataFrame, self.FieldInfo, self.CustomEvents)
        self.TestSuite.addTest(bridgeTestCase)
        self.TestSuite.addTest(commonTestCase)

    def run_tests(self) -> None:
        self.TestRunner(self.TestSuite)

    #endregion


from pyspark.sql.types import ShortType, IntegerType, LongType, FloatType, DoubleType, DecimalType, StringType
from re import search

class TestUtilities(object):

    #region Static Methods

    @staticmethod
    def check_pascal_case(s: str) -> bool:

        if not s:
            raise ValueError("s cannot be None or empty...")

        pascalPattern = r"^[A-Z][a-z]+(?:[A-Z][a-z]+)*$"
        searchResult = search(pascalPattern, s)
    
        if searchResult:
            return True
        else:
            return False

    @staticmethod
    def check_dimension_key_name(columnName: str) -> bool:

        if not columnName:
            raise ValueError("column name cannot be None or empty...")

        dimensionPrefix = "DIM_"
        dimensionSuffix = "Id"
    
        prefix = columnName[:4]
        suffix = columnName[-2:]

        return (prefix == dimensionPrefix and suffix == dimensionSuffix)

    @staticmethod
    def check_dimension_key_type(columnType: type) -> bool:

        if not columnType:
            raise ValueError("column type cannot be None...")
    
        dimensionType = IntegerType()

        return columnType == dimensionType

    @staticmethod
    def check_dimension_key_name_and_type(columnName: str, columnType: type) -> bool:

        if not columnName:
            raise ValueError("column name cannot be None or empty...")

        if not columnType:
            raise ValueError("column type cannot be None...")

        nameWellFormed = TestUtilities.check_dimension_key_name(columnName)
        tyepWellFormed = TestUtilities.check_dimension_key_type(columnType)

        return (nameWellFormed and tyepWellFormed)

    @staticmethod
    def check_flag_name(columnName: str) -> bool:

        if not columnName:
            raise ValueError("column name cannot be None or empty...")

        flagSuffix = "Flag"

        suffix = columnName[-4:]

        return suffix == flagSuffix

    @staticmethod
    def check_bad_flag_name(columnName: str) -> bool:

        if not columnName:
            raise ValueError("column name cannot be None or empty...")

        badFlagSuffix = "flag"

        suffix = columnName[-4:]

        return badFlagSuffix == suffix

    @staticmethod
    def check_flag_values(distinctColumnValues: [str]) -> bool:

        if not distinctColumnValues:
            raise ValueError("column values cannot be None or empty...")
    
        flagValues = set("Yes", "No", "Unknown", "N/A")

        for value in distinctColumnValues:
            if value not in flagValues:
                return False
    
        return True

    @staticmethod
    def check_flag_type(columnType: type) -> bool:

        if not columnType:
            raise ValueError("column type cannot be None...")

        flagType = StringType()

        return columnType == flagType

    @staticmethod
    def check_fact_type(columnType: type) -> bool:

        if not columnType:
            raise ValueError("column type cannot be None...")

        factTypes = set(ShortType(), IntegerType(), LongType(), FloatType(), DoubleType(), DecimalType())

        return (columnType in factTypes)

    #endregion


import json
import base64
import requests
import time
import uuid

class Databricks(object):

    #region Constructor

    def __init__(self, token: str, domain: str, clusterId: str = None):

        if not token:
            raise ValueError("token cannot be None or empty...")

        if not domain:
            raise ValueError("domain cannot be None or empty...")

        self.__domain = domain
        self.__token = token
        self.__uri = "https://{}/api/2.0".format(self.__domain)
        self.__header = { "Authorization", "Bearer {}".format(self.__token) }
        self.__clusterId = clusterId
    
    #endregion

    #region Class Methods

    def build_uri(self, endpoint: str, extension: str) -> str:

        if not endpoint:
            raise ValueError("endpoint cannot be None or empty...")

        if not extension:
            raise ValueError("extension cannot be None or empty...")

        uri = "{}/{}/{}".format(self.__uri, endpoint, extension)

        return uri

    def read_notebook(self, path: str, format: str = "SOURCE", direct: bool = True) -> str:

        if not path:
            raise ValueError("path cannot be None or empty...")

        endpoint = "workspace"
        extension = "export"

        uri = self.build_uri(endpoint, extension)

        body = {}
        
        body["path"] = path
        body["format"] = format
        body["direct_download"] = direct

        try:
            response = requests.get(uri, headers = self.__header, json = body)
            notebook = response.content.decode("ascii")
            return notebook
        except Exception as ex:
            raise ex

    def execute_notebook(self, path: str, existingClusterId: str = None, runName: str = None, timeOut: int = 3600) -> int:
        
        existingClusterId = existingClusterId if existingClusterId != None else self.__clusterId

        if not path:
            raise ValueError("path cannot be None or empty...")

        if not existingClusterId:
            raise ValueError("existing cluster id cannot be None or empty...")
        
        endpoint = "jobs"
        extension = "runs/submit"

        uri = self.build_uri(endpoint, extension)
        
        body = {}

        body["run_name"] = runName if runName != None else uuid.uuid4()
        body["existing_cluster_id"] = existingClusterId
        body["timeout_seconds"] = timeOut
        
        body["notebook_task"] = {}
        body["notebook_task"]["notebook_path"] = path

        try:
            response = requests.post(uri, headers = self.__header, json = body)
            runId = json.loads(response.content)["run_id"]
            return runId
        except Exception as ex:
            raise ex

    def get_run_state(self, runId: str) -> str:

        if not runId:
            raise ValueError("runId cannot be None or empty...")
        
        endpoint = "jobs"
        extension = "runs/get-output"

        uri = self.build_uri(endpoint, extension)
        
        body = {}

        body["run_id"] = runId

        try:
            response = requests.post(uri, headers = self.__header, json = body)
            # this might be limited to notebooks only, need to accomodate jobs and other executable structures if needed
            state = json.loads(response.content)["notebook_output"]["result"] 
            return state
        except Exception as ex:
            raise ex

    def get_run_uri(self, runId: str) -> str:

        if not runId:
            raise ValueError("run id cannot be None or empty...")

        endpoint = "jobs"
        extension = "runs/get"

        uri = self.build_uri(endpoint, extension)

        body = {}

        body["run_id"] = runId

        try:
            response = requests.get(uri, headers = self.__header, json = body)
            runUri = json.loads(response)["run_page_url"]
            return runUri
        except Exception as ex:
            raise ex

    def get_run_execution_time(self, runId: str) -> int:

        if not runId:
            raise ValueError("run id cannot be None or empty...")

        endpoint = "jobs"
        extension = "runs/get"

        uri = self.build_uri(endpoint, extension)

        body = {}
        body["run_id"] = runId

        try:
            response = requests.get(uri, headers = self.__header, json = body)
            runUri = json.loads(response)["execution_duration"]
            return runUri
        except Exception as ex:
            raise ex

    def get_notebook_details(self, runId: str) -> {}:

        if not runId:
            raise ValueError("run id cannot be None or empty...")

        endpoint = "jobs"
        extension = "runs/get"

        uri = self.build_uri(endpoint, extension)

        body = {}
        body["run_id"] = runId

        try:
            response = requests.get(uri, headers = self.__header, json = body)
            notebookDetails = json.loads(response)
            return notebookDetails
        except Exception as ex:
            raise ex

    def get_cluster_state(self, clusterId: str) -> str:

        clusterId = clusterId if clusterId != None else self.__clusterId

        if not clusterId:
            raise ValueError("clusterId cannot be None or empty...")
        
        endpoint = "clusters"
        extension = "get"

        uri = self.build_uri(endpoint, extension)
        
        body = {}

        body["cluster_id"] = runId

        try:
            response = requests.post(uri, headers = self.__header, json = body)
            state = json.loads(response.content)["state"] # need a wrapper class for state, enum, or const
            return state
        except Exception as ex:
            raise ex

    def start_cluster(self, clusterId: str = None) -> bool:

        clusterId = clusterId if clusterId != None else self.__clusterId

        if not clusterId:
            raise ValueError("cluster id cannot be None or empty...")

        state = self.get_cluster_state(clusterId)

        if state == "ERROR":
            raise Exception("cluster is in a state of error, please check cluster health with your admin...")

        if state == "TERMINATING":
            raise Exception("cluster is currently shutting down, cannot start cluster...")

        if state in ["RUNNING", "RESIZING", "RESTARTING"]:
            return True

        endpoint = "clusters"
        extension = "start"

        uri = self.build_uri(endpoint, extension)

        body = {}

        body["cluster_id"] = clusterId

        try:
            response = requets.post(uri, headers = self.__header, json = body)
            
            state = self.get_cluster_state(clusterId)

            while state == "PENDING":
                time.sleep(1)
                state = self.get_cluster_state(clusterId)

            if state == "Error":
                return False
            else:
                return True

        except Exception as ex:
            raise ex

    def terminate_cluster() -> bool:
        raise NotImplementedError("reserved by @phbennet")

    def create_cluster() -> bool:
        raise NotImplementedError("reserved by @phbennet")

    def create_notebook() -> bool:
        raise NotImplementedError("reserved by @phbennet")

    #endregion



from Microsoft.Finance.Common.DataBricks import Databricks
from Microsoft.Finance.Common.Spark import Spark
from Microsoft.Finance.DataStandards.Utilities.TestUtilities import TestUtilities
from Microsoft.Finance.DataStandards.Utilities.TestObject import TestObject
from pyspark.sql.session import SparkSession

class DeploymentMatrix(object):

    def __init__(self, sparkSession: SparkSession, token: str, domain: str, clusterId: str):

        if not sparkSession:
            raise ValueError("spark session cannot be None...")

        if not token:
            raise ValueError("token cannot be None or empty...")

        if not domain:
            raise ValueError("domain cannot be None or empty...")

        if not clusterId:
            raise ValueError("cluster id cannot be None or empty...")

        self.__spark = Spark(sparkSession)
        self.__databricks = Databricks(token, domain, clusterId)

    def generate_standard_matrix(self, notebookPath: str, objects: [TestObject]) -> str:

        results = {}

        results["notebook_name"] = notebookPath

        # set rc for each object
        for obj in objects:
            obj.set_row_count_before(self.__spark)

        # execute notebook
        runId = self.__databricks.execute_notebook(notebookPath)
        executionTime = self.__databricks.get_run_execution_time(runId)

        results["execution_time"] = executionTime

        # set rc for each object
        for obj in objects:
            obj.set_row_count_after(self.__spark)

        # run tests
        for obj in objects:
            if obj.Type == "dimension":
                pass
            elif obj.Type == "fact":
                pass
            else:
                pass
        pass

    def generate_comprehensive_matrix(self, notebookPath: str, objects: [TestObject]) -> str:
        raise NotImplementedError("reserved by @phbennet")


from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from typing import Set

class Spark(object):

    #region Constructor
    
    def __init__(self, sparkSession):
        self.__sparkSession = sparkSession
    
    #endregion

    #region Static Methods

    @staticmethod
    def get_single_row_result(df: DataFrame) -> object:

        if not df:
            raise ValueError("data frame cannot be None...")
    
        firstColumnName = df.schema.fields[0].name
        result = df.select(firstColumnName)
    
        return result

    @staticmethod
    def get_distinct_column_values_from_df(df: DataFrame, columnName: str) -> Set:

        if not df:
            raise ValueError("data frame cannot be None...")

        if not columnName:
            raise ValueError("column name cannot be None or empty...")

        values = set([row[0] for row in df.select(columnName).distinct().collect()])

        return values

    @staticmethod
    def get_field_info(df: DataFrame) -> {(str, type)}:

        if not df:
            raise ValueError("data frame cannot be None...")

        fieldInfo = { (field.name, field.dataType) for field in df.schema.fields }

        return fieldInfo

    #endregion

    #region Class Methods

    def get_databases(self) -> [str]:
        
        query = "show databases"
        df = spark.sql(query)
        dbs = [db[0] for db in df.select("databaseName").distinct().collect()]
        
        return dbs

    def get_tables(self, databaseName: str) -> [str]:
        
        query = "show tables in {}".format(databaseName)
        df = self.evaluate_query(query)
        tables = [row[0] for row in df.select("tableName").distinct().collect()]
        
        return tables

    def get_tables_by_database(self) -> {str: [str]}:
        
        tablesByDb = {}
        dbs = self.get_databases()
        
        for db in dbs:
            tables = self.get_tables(db)
            tablesByDb[db] = tables
        
        return tablesByDb

    def evaluate_query(self, query: str) -> DataFrame:

        if not query:
            raise ValueError("query cannot be None or empty...")
        
        try:
            df = self.__sparkSession.sql(query)
            return df
        except Exception as ex:
            raise ex

    def get_table_or_view(self, databaseName: str, tableName: str) -> DataFrame:

        if not databaseName:
            raise ValueError("database name cannot be None or empty...")

        if not tableName:
            raise ValueError("table name cannot be None or empty...")

        tableQuery = "select * from {}.{}".format(databaseName, tableName)

        df = self.evaluate_query(tableQuery)

        return df

    def count_table_or_view(self, databaseName: str, tableName: str) -> int:

        if not databaseName:
            raise ValueError("database name cannot be None or empty...")

        if not tableName:
            raise ValueError("table name cannot be None or empty...")

        countQuery = "select count(1) from {}.{}".format(databaseName, tableName)
        
        df = self.evaluate_query(countQuery)

        count = self.get_single_row_result(df)

        return count

    def count_distinct_column_values(self, databaseName: str, tableName: str, columnName: str) -> int:

        if not databaseName:
            raise ValueError("database name cannot be None or empty...")

        if not tableName:
            raise ValueError("table name cannot be None or empty...")

        if not columnName:
            raise ValueError("column name cannot be None or empty...")

        countQuery = "select count(1) from (select distinct {} from {}.{})".format(columnName, databaseName, tableName)
    
        df = self.evaluate_query(countQuery)
        
        count = self.get_single_row_result(df)

        return count
    
    def get_distinct_column_values(self, databaseName: str, tableName: str, columnName: str) -> Set:

        if not databaseName:
            raise ValueError("database name cannot be None or empty...")

        if not tableName:
            raise ValueError("table name cannot be None or empty...")

        if not columnName:
            raise ValueError("column name cannot be None or empty...")

        columnQuery = "select {} from {}.{}".format(columnName, databaseName, tableName)

        df = self.evaluate_query(columnQuery)

        distinctValues = self.get_distinct_column_values_from_df(df, columnName)

        return distinctValues

    def count_query(self, query: str) -> int:

        if not query:
            raise ValueError("query cannot be None or empty...")

        countQuery = "select count(1) from ( {} )".format(query)

        df = self.evaluate_query(countQuery)

        count = self.get_single_row_result(df)

        return count

    def get_field_info_from_query(self, query: str) -> {(str, type)}:

        if not query:
            raise ValueError("query cannot be None or empty...")

        df = self.evaluate_query(query)
        
        fieldInfo = self.get_field_info(df)
        
        return fieldInfo

    def get_field_info_table_or_view(self, databaseName: str, tableName: str) -> {(str, type)}:

        if not databaseName:
            raise ValueError("database name cannot be None or empty...")

        if not tableName:
            raise ValueError("table name cannot be None or empty...")

        df = self.get_table_or_view(databaseName, tableName)
        
        fieldInfo = self.get_field_info(df)

        return fieldInfo

    #endregion


from inspect import stack

class Utilities(object):

    #region Static Methods
    
    @staticmethod
    def get_function_caller_name():
        
        callerName = stack()[1][3]

        return callerName
    
    #endregion



from Microsoft.Finance.DataStandards.UnitTest.TestCase.CommonTestCase import CommonTestCase
from Microsoft.Finance.DataStandards.UnitTest.TestCase.DimensionTestCase import DimensionTestCase
from Microsoft.Finance.DataStandards.UnitTest.TestSuite.BaseTestSuite import BaseTestSuite
from pyspark.sql.session import SparkSession

from unittest import TestLoader, TestSuite

class DimensionTestSuite(BaseTestSuite):

    def __init__(self, sparkSession: SparkSession, databaseName: str, tableName: str) -> None:
        super(DimensionTestSuite, self).__init__(sparkSession, databaseName, tableName)

    def queue_tests(self) -> None:

        loader = TestLoader()

        names = loader.getTestCaseNames(DimensionTestCase)

        for name in names:
            self.TestSuite.addTest(DimensionTestCase(self.SparkUtility, self.FieldInfo, self.DatabaseName, self.TableName, self.TableCount, self.CustomEvents, name))

        names = loader.getTestCaseNames(CommonTestCase)

        for name in names:
            self.TestSuite.addTest(CommonTestCase(self.TableDataFrame, self.FieldInfo, self.CustomEvents, name))


from Microsoft.Finance.Common.Spark import Spark
from Microsoft.Finance.Common.Utilities import Utilities 
from Microsoft.Finance.DataStandards.Utilities.TestUtilities import TestUtilities
from Microsoft.Finance.DataStandards.Utilities.TestEvent import TestEvent
from pyspark.sql.dataframe import DataFrame

import unittest

class CommonTestCase(unittest.TestCase):

    #region Constructor

    def __init__(self, tableDataFrame: DataFrame, fieldInfo: {(str, type)}, customEvents: [TestEvent], methodName: str = "runTest"):
        super(CommonTestCase, self).__init__(methodName)
        self.__tableDataFrame = tableDataFrame
        self.__fieldInfo = fieldInfo
        self.__customEvents = customEvents

    #endregion

    #region Class Methods

    def test_pascal_casing(self) -> None:

        '''
        desc
            Conventions for fact, dimension, staging, bridge tables
                *. Column names should use Pascal Casing and should not have spaces
        '''

        event = TestEvent(Utilities.get_function_caller_name())

        eventCode1 = "pascal-case-columns"
        eventCode2 = "not-pascal-case-columns"
        eventCode3 = "dimension-keys"

        eventResults1 = []
        eventResults2 = []
        eventResults3 = []

        event.start()

        columnNames = [field[0] for field in self.__fieldInfo]

        for columnName in columnNames:
            if TestUtilities.check_dimension_key_name(columnName):
                eventResults3.append(columnName)
            else:
                if TestUtilities.check_pascal_case(columnName):
                    eventResults1.append(columnName)
                else:
                    eventResults2.append(columnName)
        
        event.end()

        event.add_note(eventCode1, eventResults1)
        event.add_note(eventCode2, eventResults2)
        event.add_note(eventCode3, eventResults3)

        self.__customEvents.append(event)

        with self.subTest():
            self.assertEquals(len(eventResults1), len(columnNames) - len(eventResults3))
        
        with self.subTest():
            self.assertEquals(len(eventResults2), 0)

    def test_flag_values(self) -> None:

        '''
        desc
            Conventions for Flags
                a. All the flag should be named as AbcFlag (examples: DynamicsFlag or CloudAddOnFlag)
                b. Possible values for Flags: Yes, No, Unknown, N/A
        '''

        event = TestEvent(Utilities.get_function_caller_name())

        eventCode1 = "well-formed-flags"
        eventCode2 = "flag-with-bad-name"
        eventCode3 = "flags-with-bad-values"
        eventCode4 = "non-flag-columns"

        eventResults1 = []
        eventResults2 = []
        eventResults3 = []
        eventResults4 = []

        event.start()

        for columnName, columnType in self.__fieldInfo:
            if TestUtilities.check_flag_name(columnName):
                if not check_flag_type(columnType):
                    eventResults3.append((columnName, columnType))
                else:
                    flagValues = Spark.get_distinct_column_values_from_df(self.__tableDataFrame, columnName)
                    if check_flag_values(flagValues):
                        eventResults1.append((columnName, columnType))
                    else:
                        eventResults3.append((columnName, columnType))
            elif TestUtilities.check_bad_flag_name(columnName):
                eventResults2.append((columnName, columnType))
            else:
                eventResults4.append((columnName, columnType))
        
        event.end()

        event.add_note(eventCode1, eventResults1)
        event.add_note(eventCode2, eventResults2)
        event.add_note(eventCode3, eventResults3)
        event.add_note(eventCode4, eventResults4)

        self.__customEvents.append(event)

        with self.subTest():
            self.assertEqual(len(eventResults2), 0)
        
        with self.subTest():
            self.assertEqual(len(eventResults3), 0)

    #endregion


from Microsoft.Finance.Common.Utilities import Utilities 
from Microsoft.Finance.DataStandards.Utilities.TestUtilities import TestUtilities
from Microsoft.Finance.DataStandards.Utilities.TestEvent import TestEvent

import unittest

class BridgeTestCase(unittest.TestCase):

    #region Constructor

    def __init__(self, fieldInfo: {(str, type)}, customEvents: [TestEvent], methodName: str = "runTest"):
        super(BridgeTestCase, self).__init__(methodName)
        self.__fieldInfo = fieldInfo
        self.__customEvents = customEvents

    #endregion

    #region Class Methods

    def test_bridge_columns_well_formed(self):
        
        '''
        desc
            Conventions for Bridge tables
                c. Output table should only contain dimension Ids and in some cases, degenerate attributes. Degenerate attributes should be added only where it is needed
        '''

        event = TestEvent(Utilities.get_function_caller_name())

        eventCode1 = "dimension-id-columns"
        eventCode2 = "degenerate-attributes"
        eventCode3 = "bad-columns"

        eventResults1 = []
        eventResults2 = []
        eventResults3 = []

        event.start()

        for columnName, columnType in self.__fieldInfo:
            if TestUtilities.check_dimension_key_name(columnName):
                if TestUtilities.check_dimension_key_type(columnType):
                    eventResults1.append((columnName, columnType))
                else:
                    eventResults3.append((columnName, columnType))
            else:
                eventResults2.append((columnName, columnType))
        
        event.end()

        event.add_note(eventCode1, eventResults1)
        event.add_note(eventCode2, eventResults2)
        event.add_note(eventCode3, eventResults3)

        self.__customEvents.append(event)

        with self.subTest():
            self.assertGreater(len(eventResults1), 0)
        
        with self.subTest():
            self.assertEquals(len(eventResults3), 0)

    #endregion


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="msftfdnepy",
    version="0.2.1",
    author="Phil Bennett",
    author_email="phbennet@microsoft.com",
    description="data testing and processing tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://microsoft.com",
    packages=setuptools.find_packages(),
    install_requires=[
        'pyspark'
    ],
    zip_safe=False,
    python_requires='>=3.0'
)


from Microsoft.Finance.DataStandards.UnitTest.TestCase.CommonTestCase import CommonTestCase
from Microsoft.Finance.DataStandards.UnitTest.TestSuite.BaseTestSuite import BaseTestSuite
from pyspark.sql.session import SparkSession

class StagingTestSuite(BaseTestSuite):

    def __init__(self, sparkSession: SparkSession, databaseName: str, tableName: str) -> None:
        super(StagingTestSuite, self).__init__(sparkSession, databaseName, tableName)

    def queue_tests(self) -> None:
        commonTestCase = CommonTestCase(self.TableDataFrame, self.FieldInfo, self.CustomEvents)
        self.TestSuite.addTest(commonTestCase)

    def run_tests(self) -> None:
        self.TestRunner.run(self.TestSuite)
